"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.analyzeReactComponent = void 0;
const type_analyzer_1 = require("@previewjs/type-analyzer");
const typescript_1 = __importDefault(require("typescript"));
const prop_types_1 = require("./prop-types");
function analyzeReactComponent(typeResolver, absoluteFilePath, componentName, signature) {
    const sourceFile = typeResolver.sourceFile(absoluteFilePath);
    let args = null;
    let propTypes = null;
    if (sourceFile) {
        args = type_analyzer_1.helpers.extractArgs(sourceFile)[componentName] || null;
        propTypes = (0, prop_types_1.detectPropTypes)(sourceFile, componentName);
    }
    const resolved = computePropsType(typeResolver, signature, propTypes);
    let providedArgs = new Set();
    if (args) {
        const argsType = typeResolver.checker.getTypeAtLocation(args);
        providedArgs = new Set(argsType.getProperties().map((prop) => prop.name));
    }
    return {
        propsType: resolved.type,
        types: { ...resolved.collected },
        providedArgs,
    };
}
exports.analyzeReactComponent = analyzeReactComponent;
function computePropsType(typeResolver, signature, propTypes) {
    if (propTypes) {
        return computePropsTypeFromPropTypes(typeResolver, propTypes);
    }
    return computePropsTypeFromSignature(typeResolver, signature);
}
function computePropsTypeFromSignature(typeResolver, signature) {
    const firstParam = signature.getParameters()[0];
    if (!firstParam) {
        return {
            type: type_analyzer_1.EMPTY_OBJECT_TYPE,
            collected: {},
        };
    }
    if (!firstParam.valueDeclaration) {
        return {
            type: type_analyzer_1.UNKNOWN_TYPE,
            collected: {},
        };
    }
    const type = typeResolver.checker.getTypeOfSymbolAtLocation(firstParam, firstParam.valueDeclaration);
    try {
        let { type: propsType, collected } = typeResolver.resolveType(type);
        [propsType] = (0, type_analyzer_1.dereferenceType)(propsType, collected, []);
        stripUnusedProps: if (propsType.kind === "object" &&
            typescript_1.default.isParameter(firstParam.valueDeclaration)) {
            if (typescript_1.default.isObjectBindingPattern(firstParam.valueDeclaration.name)) {
                const bindingPattern = firstParam.valueDeclaration.name;
                const usedProps = new Set();
                // TODO: Integrate this into Solid plugin as well.
                const propsWithDefault = new Set();
                for (const element of bindingPattern.elements) {
                    if (element.dotDotDotToken) {
                        break stripUnusedProps;
                    }
                    const elementName = element.propertyName || element.name;
                    if (!typescript_1.default.isIdentifier(elementName)) {
                        break stripUnusedProps;
                    }
                    usedProps.add(elementName.text);
                    if (element.initializer) {
                        propsWithDefault.add(elementName.text);
                    }
                }
                propsType = (0, type_analyzer_1.objectType)(Object.fromEntries(Object.entries(propsType.fields)
                    .filter(([key]) => usedProps.has(key))
                    .map(([key, type]) => [
                    key,
                    (0, type_analyzer_1.maybeOptionalType)(type, propsWithDefault.has(key)),
                ])));
            }
        }
        return { type: propsType, collected };
    }
    catch (e) {
        console.warn(`Unable to resolve props type for ${typeResolver.checker.typeToString(type)}`, e);
    }
    return {
        type: type_analyzer_1.UNKNOWN_TYPE,
        collected: {},
    };
}
function computePropsTypeFromPropTypes(typeResolver, propTypes) {
    const type = typeResolver.checker.getTypeAtLocation(propTypes);
    const fields = {};
    let collected = {};
    for (const property of type.getProperties()) {
        fields[property.name] = (() => {
            if (!property.valueDeclaration) {
                return type_analyzer_1.UNKNOWN_TYPE;
            }
            const propertyType = typeResolver.checker.getTypeAtLocation(property.valueDeclaration);
            const typeArguments = typeResolver.resolveTypeArguments(propertyType);
            const fieldType = typeArguments.types[0];
            if (fieldType) {
                collected = {
                    ...collected,
                    ...typeArguments.collected,
                };
                return (0, type_analyzer_1.maybeOptionalType)(fieldType, propertyType.symbol.name === "Requireable");
            }
            return type_analyzer_1.UNKNOWN_TYPE;
        })();
    }
    const resolved = (0, type_analyzer_1.objectType)(fields);
    return {
        type: resolved,
        collected: (0, type_analyzer_1.stripUnusedTypes)(collected, resolved),
    };
}
