"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const type_analyzer_1 = require("@previewjs/type-analyzer");
const vfs_1 = require("@previewjs/vfs");
const path_1 = __importDefault(require("path"));
const vitest_1 = require("vitest");
const _1 = require(".");
const special_types_1 = require("./special-types");
const ROOT_DIR_PATH = path_1.default.join(__dirname, "virtual");
const MAIN_FILE = path_1.default.join(ROOT_DIR_PATH, "App.tsx");
const EMPTY_SET = new Set();
(0, vitest_1.describe)("analyzeReactComponent", () => {
    let memoryReader;
    let typeAnalyzer;
    let frameworkPlugin;
    (0, vitest_1.beforeEach)(async () => {
        memoryReader = (0, vfs_1.createMemoryReader)();
        frameworkPlugin = await _1.reactFrameworkPlugin.create();
        typeAnalyzer = (0, type_analyzer_1.createTypeAnalyzer)({
            rootDirPath: ROOT_DIR_PATH,
            reader: (0, vfs_1.createStackedReader)([
                memoryReader,
                (0, vfs_1.createFileSystemReader)({
                    watch: false,
                }), // required for TypeScript libs, e.g. Promise
            ]),
            tsCompilerOptions: frameworkPlugin.tsCompilerOptions,
            specialTypes: special_types_1.REACT_SPECIAL_TYPES,
        });
    });
    (0, vitest_1.afterEach)(() => {
        typeAnalyzer.dispose();
    });
    (0, vitest_1.test)("local component with named export", async () => {
        (0, vitest_1.expect)(await analyze(`
function A() {
  return <div>Hello, World!</div>;
}

export { A }
`, "A")).toEqual({
            propsType: type_analyzer_1.EMPTY_OBJECT_TYPE,
            providedArgs: EMPTY_SET,
            types: {},
        });
    });
    (0, vitest_1.test)("local component with named aliased export", async () => {
        (0, vitest_1.expect)(await analyze(`
function A() {
  return <div>Hello, World!</div>;
}

export { A as B }
`, "A")).toEqual({
            propsType: type_analyzer_1.EMPTY_OBJECT_TYPE,
            providedArgs: EMPTY_SET,
            types: {},
        });
    });
    (0, vitest_1.test)("local component with default export", async () => {
        (0, vitest_1.expect)(await analyze(`
export function A() {
  return <div>Hello, World!</div>;
}

export default A
`, "A")).toEqual({
            propsType: type_analyzer_1.EMPTY_OBJECT_TYPE,
            providedArgs: EMPTY_SET,
            types: {},
        });
    });
    (0, vitest_1.test)("declared function with empty props", async () => {
        (0, vitest_1.expect)(await analyze(`
export function A() {
  return <div>Hello, World!</div>;
}
`, "A")).toEqual({
            propsType: type_analyzer_1.EMPTY_OBJECT_TYPE,
            providedArgs: EMPTY_SET,
            types: {},
        });
    });
    (0, vitest_1.test)("declared function with typed props parameter", async () => {
        (0, vitest_1.expect)(await analyze(`
  export function A(props: { foo: string }) {
    return <div>Hello, World!</div>;
  };
  `, "A")).toEqual({
            propsType: (0, type_analyzer_1.objectType)({
                foo: type_analyzer_1.STRING_TYPE,
            }),
            providedArgs: EMPTY_SET,
            types: {},
        });
    });
    (0, vitest_1.test)("declared function with type alias props parameter", async () => {
        (0, vitest_1.expect)(await analyze(`

  type SomeProps = {
    foo: string
  };

  export function A(props: SomeProps) {
    return <div>Hello, World!</div>;
  };
  `, "A")).toEqual({
            propsType: (0, type_analyzer_1.objectType)({ foo: type_analyzer_1.STRING_TYPE }),
            providedArgs: EMPTY_SET,
            types: {
                "App.tsx:SomeProps": {
                    type: (0, type_analyzer_1.objectType)({ foo: type_analyzer_1.STRING_TYPE }),
                    parameters: {},
                },
            },
        });
    });
    (0, vitest_1.test)("constant function with empty props", async () => {
        (0, vitest_1.expect)(await analyze(`
export const A = () => {
  return <div>Hello, World!</div>;
}
`, "A")).toEqual({
            propsType: type_analyzer_1.EMPTY_OBJECT_TYPE,
            providedArgs: EMPTY_SET,
            types: {},
        });
    });
    (0, vitest_1.test)("constant function with typed props parameter", async () => {
        (0, vitest_1.expect)(await analyze(`
export const A = (props: { foo: string }) => {
  return <div>Hello, World!</div>;
};
`, "A")).toEqual({
            propsType: (0, type_analyzer_1.objectType)({
                foo: type_analyzer_1.STRING_TYPE,
            }),
            providedArgs: EMPTY_SET,
            types: {},
        });
    });
    (0, vitest_1.test)("constant function with complex typed props parameter", async () => {
        (0, vitest_1.expect)(await analyze(`
export const A = (props: { currentTab: PanelTab, tabs: PanelTab[] }) => {
  return <div>Hello, World!</div>;
};

interface PanelTab {
  label: string;
  key: string;
  notificationCount: number;
  panel: React.ReactNode;
}
`, "A")).toEqual({
            propsType: (0, type_analyzer_1.objectType)({
                currentTab: (0, type_analyzer_1.namedType)("App.tsx:PanelTab"),
                tabs: (0, type_analyzer_1.arrayType)((0, type_analyzer_1.namedType)("App.tsx:PanelTab")),
            }),
            providedArgs: EMPTY_SET,
            types: {
                ["App.tsx:PanelTab"]: {
                    type: (0, type_analyzer_1.objectType)({
                        label: type_analyzer_1.STRING_TYPE,
                        key: type_analyzer_1.STRING_TYPE,
                        notificationCount: type_analyzer_1.NUMBER_TYPE,
                        panel: type_analyzer_1.NODE_TYPE,
                    }),
                    parameters: {},
                },
            },
        });
    });
    (0, vitest_1.test)("constant function with FunctionComponent type and no parameter", async () => {
        (0, vitest_1.expect)(await analyze(`
import { FunctionComponent } from 'react';

export const A: FunctionComponent<{ foo: string }> = (props) => {
  return <div>Hello, World!</div>;
};
`, "A")).toEqual({
            propsType: (0, type_analyzer_1.objectType)({
                foo: type_analyzer_1.STRING_TYPE,
            }),
            providedArgs: EMPTY_SET,
            types: {},
        });
    });
    (0, vitest_1.test)("constant function with FunctionComponent type and a parameter", async () => {
        (0, vitest_1.expect)(await analyze(`
import { FunctionComponent } from 'react';

export const A: FunctionComponent<{ foo: string }> = (props) => {
  return <div>Hello, {foo}!</div>;
};
`, "A")).toEqual({
            propsType: (0, type_analyzer_1.objectType)({
                foo: type_analyzer_1.STRING_TYPE,
            }),
            providedArgs: EMPTY_SET,
            types: {},
        });
    });
    (0, vitest_1.test)("constant function with FC type alias and no parameter", async () => {
        (0, vitest_1.expect)(await analyze(`
import React from 'react';

export const A: React.FC<{ foo: string }> = () => {
  return <div>Hello, World!</div>;
};
`, "A")).toEqual({
            propsType: (0, type_analyzer_1.objectType)({}),
            providedArgs: EMPTY_SET,
            types: {},
        });
    });
    (0, vitest_1.test)("constant function with FC type alias and a parameter", async () => {
        (0, vitest_1.expect)(await analyze(`
import React from 'react';

export const A: React.FC<{ foo: string }> = (props) => {
  return <div>Hello, World!</div>;
};
`, "A")).toEqual({
            propsType: (0, type_analyzer_1.objectType)({
                foo: type_analyzer_1.STRING_TYPE,
            }),
            providedArgs: EMPTY_SET,
            types: {},
        });
    });
    (0, vitest_1.test)("constant function with typed props but only some props", async () => {
        (0, vitest_1.expect)(await analyze(`
import React from 'react';

export const A: React.FC<Props> = ({ a: foo, c, d = "test" }) => {
  return <div>Hello, World!</div>;
};

type Props = {
  a: string;
  b: string;
  c: string;
  d: string;
}
`, "A")).toEqual({
            propsType: (0, type_analyzer_1.objectType)({
                a: type_analyzer_1.STRING_TYPE,
                c: type_analyzer_1.STRING_TYPE,
                d: (0, type_analyzer_1.optionalType)(type_analyzer_1.STRING_TYPE),
            }),
            providedArgs: EMPTY_SET,
            types: {
                "App.tsx:Props": {
                    type: (0, type_analyzer_1.objectType)({
                        a: type_analyzer_1.STRING_TYPE,
                        b: type_analyzer_1.STRING_TYPE,
                        c: type_analyzer_1.STRING_TYPE,
                        d: type_analyzer_1.STRING_TYPE,
                    }),
                    parameters: {},
                },
            },
        });
    });
    (0, vitest_1.test)("constant function with typed props using rest props", async () => {
        (0, vitest_1.expect)(await analyze(`
import React from 'react';

export const A: React.FC<Props> = ({ a, ...rest }) => {
  return <div>Hello, World!</div>;
};

type Props = {
  a: string;
  b: string;
  c: string;
}
`, "A")).toEqual({
            propsType: (0, type_analyzer_1.objectType)({
                a: type_analyzer_1.STRING_TYPE,
                b: type_analyzer_1.STRING_TYPE,
                c: type_analyzer_1.STRING_TYPE,
            }),
            providedArgs: EMPTY_SET,
            types: {
                "App.tsx:Props": {
                    type: (0, type_analyzer_1.objectType)({
                        a: type_analyzer_1.STRING_TYPE,
                        b: type_analyzer_1.STRING_TYPE,
                        c: type_analyzer_1.STRING_TYPE,
                    }),
                    parameters: {},
                },
            },
        });
    });
    (0, vitest_1.test)("PureComponent subclass with no props", async () => {
        (0, vitest_1.expect)(await analyze(`
import { PureComponent } from 'react';

export class A extends PureComponent {}
`, "A")).toEqual({
            propsType: type_analyzer_1.EMPTY_OBJECT_TYPE,
            providedArgs: EMPTY_SET,
            types: {},
        });
    });
    (0, vitest_1.test)("PureComponent subclass with explicit props type", async () => {
        (0, vitest_1.expect)(await analyze(`
import { PureComponent } from 'react';

export class A extends PureComponent<{foo: string}> {}
`, "A")).toEqual({
            propsType: (0, type_analyzer_1.objectType)({
                foo: type_analyzer_1.STRING_TYPE,
            }),
            providedArgs: EMPTY_SET,
            types: {},
        });
    });
    (0, vitest_1.test)("Storybook args support", async () => {
        (0, vitest_1.expect)(await analyze(`
import React from 'react';

export const A: React.FC<{ foo: string, bar: string }> = (props) => {
  return <div>{foo}</div>;
};
A.args = {
  foo: "Hello, World!"
};
`, "A")).toEqual({
            propsType: (0, type_analyzer_1.objectType)({
                foo: type_analyzer_1.STRING_TYPE,
                bar: type_analyzer_1.STRING_TYPE,
            }),
            providedArgs: new Set(["foo"]),
            types: {},
        });
    });
    (0, vitest_1.test)("PropsType support", async () => {
        (0, vitest_1.expect)(await analyze(`
import React from 'react';
import PropTypes from 'prop-types';

export const A = ({ user, onLogin, onLogout, onCreateAccount }) => {
  return <div>Hello, World!</div>;
};
A.propTypes = {
  user: PropTypes.shape({
    foo: PropTypes.string,
    bar: PropTypes.string.isRequired,
  }).isRequired,
  impersonate: PropTypes.shape({
    foo: PropTypes.string.isRequired,
  }),
  onLogin: PropTypes.func.isRequired,
  onLogout: PropTypes.func,
};
`, "A")).toEqual({
            propsType: (0, type_analyzer_1.objectType)({
                user: (0, type_analyzer_1.objectType)({
                    foo: (0, type_analyzer_1.optionalType)((0, type_analyzer_1.unionType)([type_analyzer_1.NULL_TYPE, type_analyzer_1.STRING_TYPE])),
                    bar: type_analyzer_1.STRING_TYPE,
                }),
                impersonate: (0, type_analyzer_1.optionalType)((0, type_analyzer_1.objectType)({
                    foo: type_analyzer_1.STRING_TYPE,
                })),
                onLogin: (0, type_analyzer_1.functionType)(type_analyzer_1.ANY_TYPE),
                onLogout: (0, type_analyzer_1.optionalType)((0, type_analyzer_1.functionType)(type_analyzer_1.ANY_TYPE)),
            }),
            providedArgs: EMPTY_SET,
            types: vitest_1.expect.anything(),
        });
    });
    async function analyze(source, componentName) {
        memoryReader.updateFile(MAIN_FILE, source);
        const component = (await frameworkPlugin.detectComponents(typeAnalyzer, [MAIN_FILE])).find((c) => c.name === componentName);
        if (!component) {
            throw new Error(`Component ${componentName} not found`);
        }
        return component.analyze();
    }
});
