"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const type_analyzer_1 = require("@previewjs/type-analyzer");
const vfs_1 = require("@previewjs/vfs");
const path_1 = __importDefault(require("path"));
const vitest_1 = require("vitest");
const _1 = require(".");
const extract_component_1 = require("./extract-component");
const MAIN_FILE = path_1.default.join(__dirname, "virtual", "App.tsx");
(0, vitest_1.describe)("extractReactComponents", () => {
    let memoryReader;
    let typeAnalyzer;
    (0, vitest_1.beforeEach)(async () => {
        memoryReader = (0, vfs_1.createMemoryReader)();
        const frameworkPlugin = await _1.reactFrameworkPlugin.create();
        typeAnalyzer = (0, type_analyzer_1.createTypeAnalyzer)({
            rootDirPath: path_1.default.join(__dirname, "virtual"),
            reader: (0, vfs_1.createStackedReader)([
                memoryReader,
                (0, vfs_1.createFileSystemReader)({
                    watch: false,
                }), // required for TypeScript libs, e.g. Promise
            ]),
            tsCompilerOptions: frameworkPlugin.tsCompilerOptions,
        });
    });
    (0, vitest_1.afterEach)(() => {
        typeAnalyzer.dispose();
    });
    (0, vitest_1.it)("detects expected components", async () => {
        (0, vitest_1.expect)(extract(`
import React, { Component, PureComponent } from "react";

function wrapComponent<P>(c: React.ComponentType<P>): React.ComponentType<P> {
  return c;
}

function DeclaredFunction() {
  return <div>Hello, World!</div>;
}

const ConstantFunction = () => <div>Hello, World!</div>;

// Note: this isn't detected as of October 2021.
const HocComponent = wrapComponent(() => <div>Hello, World!</div>);

class ClassComponent1 extends React.Component {}

class ClassComponent2 extends Component {}

const ForwardRef = React.forwardRef((props, ref) => <ConstantFunction {...props} />);

const NextComponent: NextPage = () => {
  return <div>Hello, World!</div>;
};

export class Pure extends PureComponent<{foo: string}> {}

export class BaseClass<T> {}

export class NotComponentClass extends BaseClass<SomeProps> {}

export const NotObjectProps = (props: () => void) => {
  return <div>Hello, World!</div>;
};

export const MissingType = (props) => {
  return <div>Hello, World!</div>;
};

export const NotComponent = () => {
  // This isn't a component.
};

export const NotComponentEither = () => {
  return "Hello";
};

export const AlsoNotAStory = {
  args: {}
};
`)).toMatchObject([
            {
                name: "DeclaredFunction",
                exported: false,
                isStory: false,
            },
            {
                name: "ConstantFunction",
                exported: false,
                isStory: false,
            },
            // Note: this isn't detected as of October 2021.
            // {
            //   name: "HocComponent",
            //   exported: false,
            //   offsets: expect.anything(),
            // },
            {
                name: "ClassComponent1",
                exported: false,
                isStory: false,
            },
            {
                name: "ClassComponent2",
                exported: false,
                isStory: false,
            },
            {
                name: "ForwardRef",
                exported: false,
                isStory: false,
            },
            {
                name: "NextComponent",
                exported: false,
                isStory: false,
            },
            {
                name: "Pure",
                exported: true,
                isStory: false,
            },
            {
                name: "NotObjectProps",
                exported: true,
                isStory: false,
            },
            {
                name: "MissingType",
                exported: true,
                isStory: false,
            },
        ]);
    });
    (0, vitest_1.it)("detects components without any React import", async () => {
        (0, vitest_1.expect)(extract(`
function DeclaredFunction() {
  return <div>Hello, World!</div>;
}

const ConstantFunction = () => <div>Hello, World!</div>;
`)).toMatchObject([
            {
                name: "DeclaredFunction",
                exported: false,
            },
            {
                name: "ConstantFunction",
                exported: false,
            },
        ]);
    });
    (0, vitest_1.it)("ignores default export of identifier", async () => {
        (0, vitest_1.expect)(extract(`
const A = () => {
  return <div>Hello, World!</div>;
};

export default A;
`)).toMatchObject([
            {
                name: "A",
                exported: true,
            },
        ]);
    });
    (0, vitest_1.it)("detects default export component (arrow function)", async () => {
        (0, vitest_1.expect)(extract(`
export default () => {
  return <div>Hello, World!</div>;
}
`)).toMatchObject([
            {
                name: "default",
                exported: true,
            },
        ]);
    });
    (0, vitest_1.it)("detects default export component (named function)", async () => {
        (0, vitest_1.expect)(extract(`
export default function test(){
  return <div>Hello, World!</div>;
}
`)).toMatchObject([
            {
                name: "test",
                exported: true,
            },
        ]);
    });
    (0, vitest_1.it)("detects default export component (anonymous function)", async () => {
        (0, vitest_1.expect)(extract(`
export default function(){
  return <div>Hello, World!</div>;
}
`)).toMatchObject([
            {
                name: "default",
                exported: true,
            },
        ]);
    });
    (0, vitest_1.it)("does not detect default export non-component", async () => {
        (0, vitest_1.expect)(extract(`
export default () => {
  return "foo";
}
`)).toMatchObject([]);
    });
    (0, vitest_1.it)("detects CSF3 stories", async () => {
        (0, vitest_1.expect)(extract(`
export default {
  component: Button
}

export const Example = {
  args: {
    label: "Hello, World!"
  }
}

export const NoArgs = {}

export function NotStory() {}
`)).toMatchObject([
            {
                name: "Example",
                exported: true,
                isStory: true,
            },
            {
                name: "NoArgs",
                exported: true,
                isStory: true,
            },
        ]);
    });
    function extract(source) {
        const rootDirPath = path_1.default.join(__dirname, "virtual");
        memoryReader.updateFile(path_1.default.join(rootDirPath, "App.tsx"), source);
        return (0, extract_component_1.extractReactComponents)(typeAnalyzer.analyze([MAIN_FILE]), MAIN_FILE);
    }
});
